@@ -11,16 +11,18 @@ class _AadHelper:
     authority_uri = None  # type: str
     token_provider = None  # type: TokenProviderBase
 
-    def __init__(self, kcsb: "KustoConnectionStringBuilder"):
+    def __init__(self, kcsb: "KustoConnectionStringBuilder", is_async: bool):
         self.kusto_uri = "{0.scheme}://{0.hostname}".format(urlparse(kcsb.data_source))
         self.username = None
 
         if kcsb.interactive_login:
-            self.token_provider = InteractiveLoginTokenProvider(self.kusto_uri, kcsb.authority_id, kcsb.login_hint, kcsb.domain_hint)
+            self.token_provider = InteractiveLoginTokenProvider(self.kusto_uri, kcsb.authority_id, kcsb.login_hint, kcsb.domain_hint, is_async=is_async)
         elif all([kcsb.aad_user_id, kcsb.password]):
-            self.token_provider = UserPassTokenProvider(self.kusto_uri, kcsb.authority_id, kcsb.aad_user_id, kcsb.password)
+            self.token_provider = UserPassTokenProvider(self.kusto_uri, kcsb.authority_id, kcsb.aad_user_id, kcsb.password, is_async=is_async)
         elif all([kcsb.application_client_id, kcsb.application_key]):
-            self.token_provider = ApplicationKeyTokenProvider(self.kusto_uri, kcsb.authority_id, kcsb.application_client_id, kcsb.application_key)
+            self.token_provider = ApplicationKeyTokenProvider(
+                self.kusto_uri, kcsb.authority_id, kcsb.application_client_id, kcsb.application_key, is_async=is_async
+            )
         elif all([kcsb.application_client_id, kcsb.application_certificate, kcsb.application_certificate_thumbprint]):
             # kcsb.application_public_certificate can be None if SNI is not used
             self.token_provider = ApplicationCertificateTokenProvider(
@@ -30,19 +32,20 @@ class _AadHelper:
                 kcsb.application_certificate,
                 kcsb.application_certificate_thumbprint,
                 kcsb.application_public_certificate,
+                is_async=is_async,
             )
         elif kcsb.msi_authentication:
-            self.token_provider = MsiTokenProvider(self.kusto_uri, kcsb.msi_parameters)
+            self.token_provider = MsiTokenProvider(self.kusto_uri, kcsb.msi_parameters, is_async=is_async)
         elif kcsb.user_token:
-            self.token_provider = BasicTokenProvider(kcsb.user_token)
+            self.token_provider = BasicTokenProvider(kcsb.user_token, is_async=is_async)
         elif kcsb.application_token:
-            self.token_provider = BasicTokenProvider(kcsb.application_token)
+            self.token_provider = BasicTokenProvider(kcsb.application_token, is_async=is_async)
         elif kcsb.az_cli:
-            self.token_provider = AzCliTokenProvider(self.kusto_uri)
-        elif kcsb.token_provider:
-            self.token_provider = CallbackTokenProvider(kcsb.token_provider)
+            self.token_provider = AzCliTokenProvider(self.kusto_uri, is_async=is_async)
+        elif kcsb.token_provider or kcsb.async_token_provider:
+            self.token_provider = CallbackTokenProvider(token_callback=kcsb.token_provider, async_token_callback=kcsb.async_token_provider, is_async=is_async)
         else:
-            self.token_provider = DeviceLoginTokenProvider(self.kusto_uri, kcsb.authority_id)
+            self.token_provider = DeviceLoginTokenProvider(self.kusto_uri, kcsb.authority_id, is_async=is_async)
 
     def acquire_authorization_header(self):
         try:
@@ -56,7 +59,7 @@ class _AadHelper:
         try:
             return _get_header_from_dict(await self.token_provider.get_token_async())
         except Exception as error:
-            kwargs = self.token_provider.context()
+            kwargs = await self.token_provider.context_async()
             kwargs["resource"] = self.kusto_uri
             raise KustoAuthenticationError(self.token_provider.name(), error, **kwargs)
 
